;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "activity-watch-mode" "1.6.0"
  "Automatic time tracking extension."
  '((emacs   "25")
    (request "0")
    (json    "0")
    (cl-lib  "0"))
  :url "https://github.com/pauldub/activity-watch-mode"
  :commit "1a950ad310cbd0511bb01744b20c7012a1c0b0e8"
  :revdesc "1a950ad310cb"
  :keywords '("calendar" "comm")
  :authors '(("Gabor Torok" . "gabor@20y.hu")
             ("Alan Hamlett" . "alan@wakatime.com"))
  :maintainers '(("Paul d'Hubert" . "paul.dhubert@ya.ru")))
